import Vue from 'vue';
import Vuex from 'vuex';
import cloneDeep from '@/modules/clone-object';
import dialplans from './modules/dialplans';
import storage from './modules/storage';
import languages from './modules/languages';
import priceLists from './modules/priceLists';
import dictionaries from './modules/dictionaries';
import pbxes from './modules/pbxes';
import features from './modules/features';
import {CONFIG} from './constants';
import {PREMIUM, BUSINESS} from '../../src/helpers/constants/license';

import AclCore from '@/modules/wms-acl';

Vue.use(Vuex);

const wmsConfig = Object.assign({}, CONFIG, cloneDeep(window.wmsConfig), {port: window.location.port});

export default new Vuex.Store({
    strict: true,
    namespaced: true,
    modules: {
        dialplans,
        dictionaries,
        languages,
        priceLists,
        pbxes,
        features,
        storage
    },
    state: {
        config: wmsConfig,
        loader: 0,
        loaderEnabled: true,
        aclCore: {}
    },
    mutations: {
        changeConfig(state, data) {
            state.config = data;
        },
        showLoader(state) {
            if (state.loaderEnabled) {
                let current = this.getters.getLoader();
                state.loader = ++current;
            }
        },
        hideLoader(state) {
            let current = this.getters.getLoader();
            if (current > 0) {
                state.loader = --current;
            }
        },
        enableLoader(state) {
            state.loaderEnabled = true;
        },
        disableLoader(state) {
            state.loaderEnabled = false;
        },
        resetLoader(state) {
            state.loader = 0;
        },
        initAcl(state, routes) {
            state.aclCore = new AclCore(routes, state.config);
        },
        recalculateAcl(state) {
            state.aclCore.recalculate(state.config);
        }
    },
    actions: {
        showLoader({commit}) {
            commit('showLoader');
        },
        hideLoader({commit}) {
            commit('hideLoader');
        },
        fetchConfig({commit, state, dispatch}) {
            dispatch('storage/init');
            this._vm.$api.PBX.Settings.Config.get()
                .then(({result}) => {
                    commit('changeConfig', Object.assign({}, state.config, result));
                    commit('recalculateAcl');
                })
                .catch(() => {
                    // TODO add errors
                    console.log('Fetch config error');
                });
        },
        initAcl({commit, getters}, routes) {
            commit('initAcl', routes);
        }
    },
    getters: {
        csrfToken: () => document.querySelector('[name="X-CSRF-Token"]').content,
        getLoader: state => () => state.loader,
        getConfig: state => () => state.config,
        isCloud: state => () => state.config.isCloud,
        isSlave: state => () => state.config.wmsNetworkRole === 'slave',
        isClient: state => () => state.config.wmsNetworkRole === 'slave',
        getUser: state => () => state.config.user,
        getNetworkInterfaces: state => () => state.config.networkInterfaces,
        acl: state => state.config.acl,
        aclCore: state => () => state.aclCore,
        isBackupsMounted: state => () => state.config.isBackupsMounted,
        isPowerCpu: state => () => state.config.hardware.isPower,
        isSubscription: state => () => state.config.license.subscription,
        isPremium: state => () => state.config.license.maximumAvailableLicenseType === PREMIUM,
        bTicinoIsInstall: state => () => {
            const bticino = state.config.integrations.find(item => item.name === 'bticino');
            return bticino.enabled;
        },
        getPort: state => () => state.config.port,
        isXBeesEnable: state => () => {
            const isRequiredLicense = [PREMIUM, BUSINESS].includes(state.config.license.maximumAvailableLicenseType);
            return state.config.license.xbees && isRequiredLicense;
        },
        isXHoppersEnable: state => () => {
            const isRequiredLicense = [PREMIUM, BUSINESS].includes(state.config.license.maximumAvailableLicenseType);
            return state.config.license.xhoppers && isRequiredLicense;
        },
        isXBeesLicenseType: state => () => state.config.license.xbees,
        isXHoppersLicenseType: state => () => state.config.license.xhoppers
    }
});
